@@ -1,6 +1,6 @@
 /* FAudio# - C# Wrapper for FAudio
  *
- * Copyright (c) 2018-2022 Ethan Lee.
+ * Copyright (c) 2018-2024 Ethan Lee.
  *
  * This software is provided 'as-is', without any express or implied warranty.
  * In no event will the authors be held liable for any damages arising from
@@ -68,7 +68,7 @@ public static class FAudio
 		}
 		return buffer;
 	}
-	
+
 	#endregion
 
 	#region FAudio API
@@ -78,8 +78,8 @@ public static class FAudio
 	public const uint FAUDIO_TARGET_VERSION = 8;
 
 	public const uint FAUDIO_ABI_VERSION =		 0;
-	public const uint FAUDIO_MAJOR_VERSION =	22;
-	public const uint FAUDIO_MINOR_VERSION =	 5;
+	public const uint FAUDIO_MAJOR_VERSION =	25;
+	public const uint FAUDIO_MINOR_VERSION =	10;
 	public const uint FAUDIO_PATCH_VERSION =	 0;
 
 	public const uint FAUDIO_COMPILED_VERSION = (
@@ -630,6 +630,11 @@ public static class FAudio
 		IntPtr voice /* FAudioVoice* */
 	);
 
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public static extern uint FAudioVoice_DestroyVoiceSafeEXT(
+		IntPtr voice /* FAudioVoice* */
+	);
+
 	/* FAudioSourceVoice Interface */
 
 	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
@@ -883,7 +888,152 @@ public static class FAudio
 
 	#region FAPO API
 
+	public const int FAPO_MIN_CHANNELS = 1,
+		FAPO_MAX_CHANNELS = 64,
+		FAPO_MIN_FRAMERATE = 1000,
+		FAPO_MAX_FRAMERATE = 200000,
+		FAPO_REGISTRATION_STRING_LENGTH = 256;
+
+	public enum FAPOBufferFlags {
+		FAPO_BUFFER_SILENT,
+		FAPO_BUFFER_VALID
+	}
+
+	[Flags]
+	public enum FAPOMiscFlags : uint {
+		FAPO_FLAG_CHANNELS_MUST_MATCH	    = 0x00000001,
+		FAPO_FLAG_FRAMERATE_MUST_MATCH	    = 0x00000002,
+		FAPO_FLAG_BITSPERSAMPLE_MUST_MATCH  = 0x00000004,
+		FAPO_FLAG_BUFFERCOUNT_MUST_MATCH    = 0x00000008,
+		FAPO_FLAG_INPLACE_SUPPORTED		    = 0x00000010,
+		FAPO_FLAG_INPLACE_REQUIRED		    = 0x00000020,
+	}
+
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate int AddRefFunc(IntPtr fapo);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+    public delegate int ReleaseFunc(IntPtr fapo);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint GetRegistrationPropertiesFunc(
+		IntPtr fapo,
+		IntPtr ppRegistrationProperties
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint IsInputFormatSupportedFunc(
+		IntPtr fapo,
+		IntPtr pOutputFormat,
+		IntPtr pRequestedInputFormat,
+		IntPtr ppSupportedInputFormat
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint IsOutputFormatSupportedFunc(
+		IntPtr fapo,
+		IntPtr pInputFormat,
+		IntPtr pRequestedOutputFormat,
+		IntPtr ppSupportedOutputFormat
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint InitializeFunc(
+		IntPtr fapo,
+		IntPtr pData,
+		uint DataByteSize
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate void ResetFunc(
+		IntPtr fapo
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint LockForProcessFunc (
+		IntPtr fapo,
+		uint InputLockedParameterCount,
+		ref FAPOLockForProcessBufferParameters pInputLockedParameters,
+		uint OutputLockedParameterCount,
+		ref FAPOLockForProcessBufferParameters pOutputLockedParameters
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate void UnlockForProcessFunc (
+		IntPtr fapo
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate void ProcessFunc (
+		IntPtr fapo,
+		uint InputProcessParameterCount,
+		ref FAPOProcessBufferParameters pInputProcessParameters,
+		uint OutputProcessParameterCount,
+		ref FAPOProcessBufferParameters pOutputProcessParameters,
+		int IsEnabled
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint CalcInputFramesFunc (
+		IntPtr fapo,
+		uint OutputFrameCount
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate uint CalcOutputFramesFunc (
+		IntPtr fapo,
+		uint InputFrameCount
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate void SetParametersFunc (
+		IntPtr fapo,
+		IntPtr pParameters,
+		uint ParameterByteSize
+	);
+	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+	public delegate void GetParametersFunc (
+		IntPtr fapo,
+		IntPtr pParameters,
+		uint ParameterByteSize
+	);
+
 	/* TODO */
+	[StructLayout(LayoutKind.Sequential)]
+	public struct FAPO {
+		public IntPtr AddRef;
+		public IntPtr Release;
+		public IntPtr GetRegistrationProperties;
+		public IntPtr IsInputFormatSupported;
+		public IntPtr IsOutputFormatSupported;
+		public IntPtr Initialize;
+		public IntPtr Reset;
+		public IntPtr LockForProcess;
+		public IntPtr UnlockForProcess;
+		public IntPtr Process;
+		public IntPtr CalcInputFrames;
+		public IntPtr CalcOutputFrames;
+		public IntPtr SetParameters;
+		public IntPtr GetParameters;
+	}
+
+	[StructLayout(LayoutKind.Sequential)]
+	public unsafe struct FAPORegistrationProperties
+	{
+		public Guid clsid;
+		public fixed char FriendlyName[FAPO_REGISTRATION_STRING_LENGTH]; /* Win32 wchar_t */
+		public fixed char CopyrightInfo[FAPO_REGISTRATION_STRING_LENGTH]; /* Win32 wchar_t */
+		public uint MajorVersion;
+		public uint MinorVersion;
+		public uint Flags;
+		public uint MinInputBufferCount;
+		public uint MaxInputBufferCount;
+		public uint MinOutputBufferCount;
+		public uint MaxOutputBufferCount;
+	};
+
+	[StructLayout(LayoutKind.Sequential)]
+	public struct FAPOLockForProcessBufferParameters
+	{
+		public IntPtr pFormat;
+		uint MaxFrameCount;
+	}
+
+	[StructLayout(LayoutKind.Sequential)]
+	public struct FAPOProcessBufferParameters
+	{
+		public IntPtr pBuffer;
+		public FAPOBufferFlags BufferFlags;
+		public uint ValidFrameCount;
+	}
 
 	#endregion
 
@@ -891,6 +1041,82 @@ public static class FAudio
 
 	/* TODO */
 
+	[StructLayout(LayoutKind.Sequential, Pack = 8)]
+	public struct FAPOBase {
+		public FAPO FAPO;
+		public IntPtr Destructor;
+		public IntPtr OnSetParameters;
+
+		IntPtr m_pRegistrationProperties;
+		IntPtr m_pfnMatrixMixFunction;
+		IntPtr m_pfl32MatrixCoefficients;
+		uint m_nSrcFormatType;
+		byte m_fIsScalarMatrix;
+		byte m_fIsLocked;
+		IntPtr m_pParameterBlocks;
+		IntPtr m_pCurrentParameters;
+		IntPtr m_pCurrentParametersInternal;
+		uint m_uCurrentParametersIndex;
+		uint m_uParameterBlockByteSize;
+		byte m_fNewerResultsReady;
+		byte m_fProducer;
+
+		/* Protected Variables */
+		int m_lReferenceCount; /* LONG */
+
+		/* Allocator callbacks, NOT part of XAPOBase spec! */
+		IntPtr pMalloc;
+		IntPtr pFree;
+		IntPtr pRealloc;
+	}
+
+	[StructLayout(LayoutKind.Sequential, Pack = 4)]
+	public struct FAudioFXCollectorState {
+		public UInt32 WriteOffset;
+	};
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public static extern uint FAudioCreateCollectorEXT(
+		out IntPtr ppApo,
+		UInt32 flags,
+		IntPtr pBuffer,
+		UInt32 bufferLength
+	);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public static extern uint FAudioCreateCollectorWithCustomAllocatorEXT(
+		out IntPtr ppApo,
+		UInt32 flags,
+		IntPtr pBuffer,
+		UInt32 bufferLength,
+		IntPtr customMalloc,
+		IntPtr customFree,
+		IntPtr customRealloc
+	);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public static extern uint CreateFAPOBase(
+		IntPtr fapo,
+		// The create operation does not copy this into its own persistent storage! This pointer must remain valid.
+		IntPtr pRegistrationProperties,
+		IntPtr pParameterBlocks,
+		uint parameterBlockByteSize,
+		byte fProducer
+	);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public static extern uint CreateFAPOBaseWithCustomAllocatorEXT(
+		IntPtr fapo,
+		// The create operation does not copy this into its own persistent storage! This pointer must remain valid.
+		IntPtr pRegistrationProperties,
+		IntPtr pParameterBlocks,
+		uint parameterBlockByteSize,
+		byte fProducer,
+		IntPtr customMalloc,
+		IntPtr customFree,
+		IntPtr customRealloc
+	);
+
 	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
 	public static extern uint FAPOBase_Release(IntPtr fapo);
 
@@ -1947,7 +2173,7 @@ public static class FAudio
 	[StructLayout(LayoutKind.Sequential, Pack = 1)]
 	public struct F3DAUDIO_DISTANCE_CURVE
 	{
-		IntPtr pPoints; /* F3DAUDIO_DISTANCE_CURVE_POINT* */
+		public IntPtr pPoints; /* F3DAUDIO_DISTANCE_CURVE_POINT* */
 		public uint PointCount;
 	}
 
@@ -2385,4 +2611,43 @@ public static class FAudio
 	);
 
 	#endregion
+
+	#region qoa
+
+	/* Because, again, why not? */
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public extern static IntPtr qoa_open_from_memory(IntPtr bytes, uint size, int free_on_close);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	private static extern unsafe IntPtr qoa_open_from_filename(
+		byte* filename
+	);
+
+	public static unsafe IntPtr qoa_open_from_filename(
+		string filename
+	) {
+		int utf8BufSize = Utf8Size(filename);
+		byte* utf8Buf = stackalloc byte[utf8BufSize];
+		return qoa_open_from_filename(
+			Utf8Encode(filename, utf8Buf, utf8BufSize)
+		);
+	}
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public extern static unsafe void qoa_attributes(IntPtr qoa, out uint channels, out uint samplerate, out uint samples_per_channel_per_frame, out uint total_samples_per_channel);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public extern static unsafe uint qoa_decode_next_frame(IntPtr qoa, short *sample_data);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public extern static unsafe void qoa_seek_frame(IntPtr qoa, int frame_index);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public extern static unsafe void qoa_decode_entire(IntPtr qoa, short *sample_data);
+
+	[DllImport(nativeLibName, CallingConvention = CallingConvention.Cdecl)]
+	public extern static unsafe void qoa_close(IntPtr qoa);
+
+	#endregion
 }
